import { GArray, Node, Object, StringName } from 'godot'

interface Predicate<T> {
  (value: T): boolean
}

export function queue_free_children<T extends Node>(node: T) {
  const children: GArray<Node> = node.get_children()
  for (const child of children) {
    child.queue_free()
  }
}

type RequiredField<T, Key extends keyof T> = Omit<T, Key> & Required<Pick<T, Key>>;

export function has_methods<T extends Object & {}, U extends [keyof T, ...Array<keyof T>]>(
  keys: U,
  obj: T
): obj is Object & RequiredField<T, U[number]> {
  return keys.every(key => obj.has_method(key as string))
}

export function has_method<T extends Object & {}>(name: StringName, obj: Object): obj is Object & RequiredField<T, keyof T> {
  return has_methods([name as keyof Object], obj)
}

export function implements_interface<T>(method_names: StringName[], obj: Object): obj is Object & T {
  return method_names.every(name => has_method(name, obj))
}

export function find_in_ancestors<T extends Node>(root: T | null | undefined, predicate: Predicate<Node>, height: number = Infinity): T | null | undefined {
  if (height < 0 || root == null) {
    return
  }

  if (predicate(root)) {
    return root
  }

  return find_in_ancestors(root.get_parent(), predicate, height - 1) as T
}

export function find_in_descendents<T extends Node>(root: T, predicate: Predicate<Node>, depth: number = Infinity): T | null | undefined {
  if (depth < 0) {
    return
  }

  if (predicate(root)) {
    return root
  }

  const count = root.get_child_count()
  for (let i = 0; i < count; i++) {
    const child = root.get_child(i)
    const found = find_in_descendents(child, predicate, depth - 1)

    if (found != null) {
      return found as T
    }
  }
}
