import { FileAccess, GArray, Node, PackedScene } from 'godot'
import AsyncResourceHandler from '../global/async_resource_handler'
import { implements_interface } from '../extension/node'
import GameState, { Savable } from './game_state'
import SaveState from './save_state'
import { Enumerable } from '../../addons/enumerable-ts/src'
import { GArrayEnumerator } from '../collection/enumerable'

interface FileHandler {
  (file: FileAccess): void
}

const open = (path: string, mode: FileAccess.ModeFlags, fn: FileHandler) => {
  const fd = FileAccess.open(path, mode)
  const result = fn(fd)
  fd.flush()
  fd.close()
  return result
}

const read = (path: string, handler: FileHandler) => open(path, FileAccess.ModeFlags.READ, handler)

const write = (path: string, handler: FileHandler) => open(path, FileAccess.ModeFlags.WRITE, handler)

const StateMethod = {
  Save: 'on_save_game',
  BeforeLoad: 'on_before_load_game',
  Load: 'on_load_game'
} as const

type StateMethod = typeof StateMethod[keyof typeof StateMethod]

export default class SaveSerializer extends Node {
  private static _save_group: string = 'save'

  private static _instance: SaveSerializer

  _ready() {
    SaveSerializer._instance = this
    this.save()
    this.load(this)
  }

  static save() {
    this._instance.save()
  }

  private _announce(method: StateMethod, ...args: any) {
    this.get_tree().call_group(
      SaveSerializer._save_group,
      method,
      ...args
    )
  }

  save() {
    write('user://save1.save', file => {
      const game_state = new GameState()
      this._announce(StateMethod.Save, game_state)
      //const json = JSON.stringify(game_state)
      file.store_var(game_state)
    })
  }

  private async _restore_node<T extends SaveState>(state: T, parent: Node) {
    const scene: PackedScene = await AsyncResourceHandler.load(state.scene, 'PackedScene')
    const node = scene.instantiate()
    parent.add_child(node)

    if (implements_interface<Savable>(Object.values(StateMethod), node)) {
      node.on_load_game(state)
    }
  }

  async load(world_root: Node) {
    read('user://save1.save', file => {
      //const state = JSON.parse(file.get_line()) as GameState
      const state = file.get_var() as GameState
      console.log(state)
      this._announce(StateMethod.BeforeLoad)
      const promises = []

      for (const obj of state.world) {
        promises.push(this._restore_node(obj, world_root))
      }

      return Promise.all(promises)
    })
  }
}

